from typing import Any, List

from statham.schema.elements.base import Element
from statham.schema.constants import Maybe, NotPassed
from statham.schema.validation import InstanceOf


class String(Element[str]):
    """JSON Schema ``"string"`` element."""

    def __init__(
        self,
        *,
        default: Maybe[str] = NotPassed(),
        const: Maybe[Any] = NotPassed(),
        enum: Maybe[List[Any]] = NotPassed(),
        # Bad name to match JSON Schema keywords.
        # pylint: disable=redefined-builtin
        format: Maybe[str] = NotPassed(),
        # pylint: enable=redefined-builtin
        pattern: Maybe[str] = NotPassed(),
        minLength: Maybe[int] = NotPassed(),
        maxLength: Maybe[int] = NotPassed(),
    ):
        self.default = default
        self.const = const
        self.enum = enum
        self.format = format
        self.pattern = pattern
        self.minLength = minLength
        self.maxLength = maxLength

    @property
    def type_validator(self):
        return InstanceOf(str)



import inspect
from functools import partial
from typing import Any, Dict, Optional

from statham.schema.elements import (
    Array,
    Boolean,
    CompositionElement,
    Element,
    Integer,
    Not,
    Nothing,
    Null,
    Number,
    String,
)
from statham.schema.elements.meta import ObjectMeta
from statham.schema.property import _Property
from statham.serializers.orderer import get_object_classes


def serialize_json(
    *elements: Element, definitions: Dict[str, Element] = None
) -> Dict[str, Any]:
    """Serialize elements to a JSON Schema dictionary.

    Object classes are included in definitions. The first element is
    the top-level schema.

    :param elements: The :class:`~statham.schema.elements.Element` objects
        to serialize.
    :param definitions: A dictionary of elements which should be members
      of the schema definitions keyword, and referenced everywhere else.
    :return: A JSON-serializable dictionary containing the JSON Schema
        for the provided element(s).
    """
    primary = elements[0]
    object_classes = get_object_classes(*elements)
    serialize = partial(
        _serialize_element, object_refs=True, definitions=definitions
    )
    schema: Dict[str, Any] = {
        **serialize(primary),
        "definitions": {
            object_class.__name__: serialize(object_class)
            for object_class in object_classes
            if object_class is not primary
        },
    }
    if definitions:
        schema["definitions"].update(
            {key: serialize(element) for key, element in definitions.items()}
        )
    if not schema["definitions"]:
        del schema["definitions"]
    return schema


def _serialize_element(
    element: Element,
    object_refs: bool = False,
    definitions: Dict[str, Any] = None,
):
    """Convert a schema element to a JSON Schema dictionary.

    This does the heavy lifting behind `serialize_json`.
    """
    if isinstance(element, Nothing):
        return False
    schema = {
        param.name: getattr(element, param.name, param.default)
        for param in inspect.signature(Element.__init__).parameters.values()
        if param.kind == param.KEYWORD_ONLY
        and getattr(element, param.name, param.default) != param.default
    }
    if not schema.get("properties", True):
        del schema["properties"]
    if "properties" in schema:
        schema["required"] = [
            prop.source or name
            for name, prop in schema["properties"].items()
            if prop.required
        ]
    if not schema.get("required", True):
        del schema["required"]
    if isinstance(element, CompositionElement):
        schema[element.mode] = element.elements
    if isinstance(element, Not):
        schema["not"] = element.element
    if type(element) in _TYPE_MAPPING:  # pylint: disable=unidiomatic-typecheck
        schema["type"] = _TYPE_MAPPING[type(element)]
    if isinstance(element, ObjectMeta):
        schema["title"] = element.__name__
    return _serialize_recursive(
        schema, object_refs=object_refs, definitions=definitions
    )


_TYPE_MAPPING = {
    Array: "array",
    Boolean: "boolean",
    Integer: "integer",
    Null: "null",
    ObjectMeta: "object",
    Number: "number",
    String: "string",
}


def _serialize_recursive(
    data: Any, object_refs: bool = False, definitions: Dict[str, Element] = None
) -> Any:
    """Recursively serialize schema elements."""
    recur = partial(
        _serialize_recursive, object_refs=object_refs, definitions=definitions
    )
    if isinstance(data, _Property):
        data = data.element
    if isinstance(data, ObjectMeta) and object_refs:
        return {"$ref": f"#/definitions/{data.__name__}"}
    if isinstance(data, Element):
        return _from_definitions(
            definitions,
            data,
            _serialize_element(
                data, object_refs=object_refs, definitions=definitions
            ),
        )
    if not isinstance(data, (list, dict)):
        return data
    if isinstance(data, list):
        return [recur(item) for item in data]
    return {key: recur(value) for key, value in data.items()}


def _from_definitions(
    definitions: Optional[Dict[str, Element]],
    element: Element,
    default: Any = None,
):
    """Check if this element is present in definitions."""
    if not definitions:
        return default
    return next(
        (
            {"$ref": f"#/definitions/{key}"}
            for key, definition in definitions.items()
            if definition == element
        ),
        default,
    )



from typing import Any, ClassVar, Optional, Tuple, Type

from statham.schema.constants import NotPassed
from statham.schema.exceptions import ValidationError


_TRUE = object()
_FALSE = object()


def replace_bool(value: Any) -> Any:
    return _TRUE if value is True else _FALSE if value is False else value


def _is_instance(value, type_args):
    """Variant of isinstance to handle booleans correctly.

    In CPython, isinstance(True, int) evaluates True.
    """
    if isinstance(value, bool):
        if bool in type_args:
            return True
        return False
    return isinstance(value, type_args)


class Validator:
    """Base validator type.

    Logic for given validation keywords is implemented in subclasses by
    overriding class variables and implementing the ``_validate()``
    method.
    """

    types: ClassVar[Optional[Tuple[Type, ...]]] = None
    """Types on which this validator applies.

    If ``None``, apply to all values.
    """

    message: ClassVar[str] = ""
    """The error message to display on validation failure.

    Can be tamplated on :attr:`Validator.keywords`.
    """

    keywords: Tuple[str, ...] = tuple()
    """Keywords which configure this validator.

    These are used to configure the validator based on the
    :class:`statham.schema.elements.Element`.
    """

    def __init__(self, *args):
        """Accepts the parameters specified by the `keywords` class variable."""
        if len(self.keywords) != len(args):
            raise TypeError(
                f"{type(self).__name__}.__init__ takes exactly "
                f"{len(self.keywords)} ({len(args)} given)"
            )
        self.params = dict(zip(self.keywords, args))

    @classmethod
    def from_element(cls, element) -> Optional["Validator"]:
        """Construct validator from an Element instance.

        Check for attributes matching keywords for this validator. If
        none are present, then return None.
        """
        params = tuple(
            getattr(element, keyword, NotPassed()) for keyword in cls.keywords
        )
        if NotPassed() in params:
            return None
        return cls(*params)

    # pylint: disable=no-self-use
    def _validate(self, _value: Any):
        """Validate a value.

        Validation logic should be added by base classes, and raise a
        bare `ValidationError` on a failure. The error message will be
        automatically generated from the `message` class variable for
        consistency.
        """
        return

    def error_message(self):
        """Generate the error message on failed validation."""
        return self.message.format(**self.params)

    def __call__(self, value: Any, property_: Any):
        """Apply the validator to a value.

        Checks that `value` has correct type for this validator, runs
        validation logic and constructs the error message on failure.

        :param value: The value to validate.
        :param property_: The enclosing property if present - used for
            error reporting.
        :raises: :class:`~statham.schema.exceptions.ValidationError` if
            :paramref:`~Validator.__call__.value` fails validation.
        """
        if self.types and not _is_instance(value, self.types):
            return
        try:
            self._validate(value)
        except ValidationError:
            raise ValidationError.from_validator(
                property_, value, self.error_message()
            )


class InstanceOf(Validator):
    """Validate the type of a value."""

    message = "Must be of type {type_names}."

    def __init__(self, *args):
        super().__init__()
        self.params["types"] = args
        names = (type_.__name__ for type_ in self.params["types"])
        self.params["type_names"] = f"({','.join(names)})"

    def _validate(self, value: Any):
        if value == NotPassed() or not self.params["types"]:
            return
        if not _is_instance(value, self.params["types"]):
            raise ValidationError


class NoMatch(Validator):
    """Don't accept any passed value.

    Used exclusively by :class:`~statham.schema.elements.base.Nothing`.
    """

    message = "Schema does not accept any values."

    def _validate(self, value: Any):
        if value is NotPassed():
            return
        raise ValidationError


class Const(Validator):
    """Validate that passed values match a constant value."""

    keywords = ("const",)
    message = "Must match constant value: {const}"

    def _validate(self, value: Any):
        aliased = replace_bool(value)
        const = replace_bool(self.params["const"])
        if aliased != const:
            raise ValidationError


class Enum(Validator):
    """Validate that passed values are members of an enumeration."""

    keywords = ("enum",)
    message = "Must be one of these values: {enum}"

    def _validate(self, value: Any):
        aliased = replace_bool(value)
        enum = list(map(replace_bool, self.params["enum"]))
        if aliased not in enum:
            raise ValidationError



from typing import Any, List, TypeVar, Union

from statham.schema.constants import Maybe, NotPassed
from statham.schema.elements.base import Element
from statham.schema.helpers import remove_duplicates
from statham.schema.validation import InstanceOf


Item = TypeVar("Item")


# pylint: disable=too-many-instance-attributes
class Array(Element[List[Item]]):  # pylint: disable=missing-param-doc
    """JSON Schema ``"array"`` element.

    :param items:
        As in :class:`statham.schema.elements.Element`, but as a required
        positional argument.
    """

    items: Union[Element[Item], List[Element]]

    def __init__(
        self,
        items: Union[Element[Item], List[Element]],
        *,
        default: Maybe[List] = NotPassed(),
        const: Maybe[Any] = NotPassed(),
        enum: Maybe[List[Any]] = NotPassed(),
        additionalItems: Union[Element, bool] = True,
        minItems: Maybe[int] = NotPassed(),
        maxItems: Maybe[int] = NotPassed(),
        uniqueItems: bool = False,
        contains: Maybe[Element] = NotPassed(),
    ):
        self.items = items
        self.default = default
        self.const = const
        self.enum = enum
        self.additionalItems = additionalItems
        self.minItems = minItems
        self.maxItems = maxItems
        self.uniqueItems = uniqueItems
        self.contains = contains

    @property
    def annotation(self) -> str:
        if not self.item_annotations:
            return "List"
        if len(self.item_annotations) == 1:
            return f"List[{self.item_annotations[0]}]"
        return f"List[Union[{', '.join(self.item_annotations)}]]"

    @property
    def item_annotations(self) -> List[str]:
        """Get a list of possible type annotations."""
        if isinstance(self.items, Element):
            return [self.items.annotation]
        annotations: List[str] = [item.annotation for item in self.items]
        if self.additionalItems is True:
            return ["Any"]
        if isinstance(self.additionalItems, Element):
            annotations.append(self.additionalItems.annotation)
        if "Any" in annotations:
            return ["Any"]
        return remove_duplicates(annotations)

    @property
    def type_validator(self):
        return InstanceOf(list)



from typing import Any, Dict, List, TypeVar, Union


JSONElement = Union[Dict[str, Any], List[Any], int, float, bool, None, str]


class NotPassed:
    """Singleton similar to NoneType.

    To distinguish between arguments not passed, and arguments passed as
    None.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Enforce singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __repr__(self) -> str:
        return "NotPassed"

    def __bool__(self) -> bool:
        return False


T = TypeVar("T")

Maybe = Union[T, NotPassed]


COMPOSITION_KEYWORDS = ("anyOf", "oneOf", "allOf", "not")


UNSUPPORTED_SCHEMA_KEYWORDS = {
    "$defs",
    "if",
    "then",
    "else",
    "unevaluatedItems",
    "unevaluatedProperties",
}



from typing import Any, List, NamedTuple, Optional, TypeVar
from typing_extensions import Literal

from statham.schema.constants import NotPassed
from statham.schema.elements.base import Element
from statham.schema.helpers import remove_duplicates
from statham.schema.exceptions import ValidationError
from statham.schema.property import _Property


# This is a type annotation.
Mode = Literal["anyOf", "oneOf", "allOf"]  # pylint: disable=invalid-name
T = TypeVar("T")


class Not(Element[T]):
    """JSON Schema ``"not"`` element.

    Element fails to validate if enclosed schema validates.

    :param element: The enclosed :class:`~statham.schema.elements.Element`.
    :param default: Inherited from :class:`~statham.schema.elements.Element`.
    """

    def __init__(self, element: Element, *, default: Any = NotPassed()):
        self.element = element
        super().__init__(default=default)

    def construct(self, value: Any, property_: _Property):
        try:
            _ = self.element(value, property_)
        except (TypeError, ValidationError):
            return value
        raise ValidationError.from_validator(
            property_, value, f"Must not match {self.element}."
        )


class CompositionElement(Element):
    """Composition Base Element.

    The "oneOf", "anyOf" and "allOf" elements share the same interface.

    :param elements: The composed :class:`~statham.schema.elements.Element`
        objects.
    :param default: Inherited from :class:`~statham.schema.elements.Element`.
    """

    mode: Mode

    def __init__(self, *elements: Element, default: Any = NotPassed()):
        super().__init__()
        self.elements = list(elements)
        if not self.elements:
            raise TypeError(f"{type(self)} requires at least one sub-schema.")
        self.default = default

    @property
    def annotation(self):
        annotations = remove_duplicates(
            elem.annotation for elem in self.elements
        )
        if len(annotations) == 1:
            return annotations[0]
        if "Any" in annotations:
            return "Any"
        joined = ", ".join(annotations)
        return f"Union[{joined}]"

    def construct(self, value: Any, property_: _Property):
        if not getattr(self, "mode", None):
            raise NotImplementedError
        return _attempt_schemas(self.elements, value, property_, mode=self.mode)


class AnyOf(CompositionElement):
    """JSON Schema ``"anyOf"`` element.

    Must match at least one of the provided schemas.

    :param elements: The composed :class:`~statham.schema.elements.Element`
        objects.
    :param default: Inherited from :class:`~statham.schema.elements.Element`.
    """

    mode: Mode = "anyOf"


class OneOf(CompositionElement):
    """JSON Schema ``"oneOf"`` element.

    Must match exactly one of the provided schemas.

    :param elements: The composed :class:`~statham.schema.elements.Element`
        objects.
    :param default: Inherited from :class:`~statham.schema.elements.Element`.
    """

    mode: Mode = "oneOf"


class AllOf(CompositionElement):
    """JSON Schema ``"allOf"`` element.

    Must match all provided schemas.

    :param elements: The composed :class:`~statham.schema.elements.Element`
        objects.
    :param default: Inherited from :class:`~statham.schema.elements.Element`.
    """

    mode: Mode = "allOf"

    @property
    def annotation(self):
        """Get type annotation for element.

        With an AllOf element, this should be the most restrictive
        type. On the assumption that the composition is possible, this
        is chosen by returning the first explicit type annotation, or the
        first union type annotation if no explicit annotations are
        present.
        """
        return next(
            (
                elem.annotation
                for elem in self.elements
                if elem.annotation != "Any"
                and not elem.annotation.startswith("Union")
            ),
            next(
                (
                    elem.annotation
                    for elem in self.elements
                    if elem.annotation != "Any"
                ),
                "Any",
            ),
        )


class Outcome(NamedTuple):

    target: Element
    result: Optional[Any] = None
    error: Optional[Exception] = None


def _attempt_schema(
    element: Element, value: Any, property_: _Property
) -> Outcome:
    """Attempt to pass an input to a schema element.

    :return: An `Outcome` object describing containing success/failure
        information and a result if successful.
    """
    try:
        return Outcome(element, result=element(value, property_), error=None)
    except (TypeError, ValidationError) as exc:
        return Outcome(element, result=None, error=exc)


def _attempt_schemas(
    elements: List[Element],
    value: Any,
    property_: _Property,
    mode: str = "anyOf",
) -> Any:
    """Attempt to instantiate a given input against many elements.

    :param elements: The elements against which to validate.
    :param property_: The enclosing property.
    :param value: The data to validate against.
    :param mode: The matching stategy to use. "allOf", "anyOf" or "oneOf".
    :return: A list of successful instantiations against `elements`.
    :raises ValidationError: if there are no matching schemas.
    :raises ValueError: if passed an invalid mode.
    """
    outcomes = [
        _attempt_schema(element, value, property_) for element in elements
    ]
    results = [outcome.result for outcome in outcomes if not outcome.error]
    errors = [outcome.error for outcome in outcomes if outcome.error]
    if not results:
        raise ValidationError.combine(
            property_, value, errors, "Does not match any accepted schema."
        )

    if mode == "anyOf":
        return results[0]
    if mode == "oneOf":
        if len(results) > 1:
            raise ValidationError.multiple_composition_match(
                [outcome.target for outcome in outcomes if not outcome.error],
                value,
            )
        return results[0]
    if mode == "allOf":
        if errors:
            raise ValidationError.combine(
                property_, value, errors, "Does not match all required schemas."
            )
        return results[0]
    raise ValueError(f"Got bad argument for `mode`: {mode}")  # pragma: no cover



from typing import Any

from statham.schema.exceptions import ValidationError
from statham.schema.validation.base import Validator


class Required(Validator):
    """Validate that object values contain all required properties."""

    types = (dict,)
    keywords = ("required",)
    message = "Must contain all required fields: {required}"

    @classmethod
    def from_element(cls, element):
        required = getattr(element, "required", None) or []
        properties = getattr(element, "properties", None)
        if properties:
            required += properties.required
        if not required:
            return None
        return Required(required)

    def _validate(self, value: Any):
        if set(self.params["required"]) - set(value):
            raise ValidationError


class AdditionalProperties(Validator):
    """Validate that prohibited properties are not included."""

    types = (dict,)
    keywords = ("__properties__",)
    message = "Must not contain unspecified properties. Accepts: {properties}"

    def error_message(self):
        return self.message.format(
            properties=set(self.params["__properties__"])
        )

    def _validate(self, value: Any):
        if self.params["__properties__"].additional:
            return
        bad_properties = {
            key for key in value if key not in self.params["__properties__"]
        }
        if bad_properties:
            raise ValidationError


class MinProperties(Validator):
    """Validate that object values contain a minimum number of properties."""

    types = (dict,)
    keywords = ("minProperties",)
    message = "Must contain at least {minProperties} properties."

    def _validate(self, value: Any):
        if len(value) < self.params["minProperties"]:
            raise ValidationError


class MaxProperties(Validator):
    """Validate that object values contain a maximum number of properties."""

    types = (dict,)
    keywords = ("maxProperties",)
    message = "Must contain at most {maxProperties} properties."

    def _validate(self, value: Any):
        if len(value) > self.params["maxProperties"]:
            raise ValidationError


class PropertyNames(Validator):
    """Validate that property names conform to a schema."""

    types = (dict,)
    keywords = ("propertyNames",)
    message = "Property names must match schema {propertyNames}"

    def _validate(self, value: Any):
        for prop_name in value:
            try:
                self.params["propertyNames"](prop_name)
            except (ValidationError, TypeError):
                raise ValidationError


class Dependencies(Validator):
    types = (dict,)
    keywords = ("dependencies",)
    message = "Must match defined dependencies: {dependencies}."

    def _validate(self, value: Any):
        for key, dep in self.params["dependencies"].items():
            if key not in value:
                continue
            if isinstance(dep, list):
                # pylint: disable=protected-access
                Required(dep)._validate(value)
            else:
                self.validate_schema_dependency(dep, value)

    @staticmethod
    def validate_schema_dependency(dependency, value):
        try:
            _ = dependency(value)
        except (TypeError, ValidationError):
            raise ValidationError



import re
from typing import Any

from statham.schema.exceptions import ValidationError
from statham.schema.validation.base import Validator
from statham.schema.validation.format import format_checker


class Pattern(Validator):
    """Validate that string values match a regular expression."""

    types = (str,)
    keywords = ("pattern",)
    message = "Must match regex pattern {pattern}."

    def error_message(self):
        return self.message.format(pattern=repr(self.params["pattern"]))

    def _validate(self, value: Any):
        if not re.search(self.params["pattern"], value):
            raise ValidationError


class MinLength(Validator):
    """Validate that string values are over a minimum length."""

    types = (str,)
    keywords = ("minLength",)
    message = "Must be at least {minLength} characters long."

    def _validate(self, value: Any):
        if len(value) < self.params["minLength"]:
            raise ValidationError


class MaxLength(Validator):
    """Validate that string values are under a maximum length."""

    types = (str,)
    keywords = ("maxLength",)
    message = "Must be at most {maxLength} characters long."

    def _validate(self, value: Any):
        if len(value) > self.params["maxLength"]:
            raise ValidationError


class Format(Validator):
    """Validate that string values match a named format.

    Additional formats may be registered via
    :func:`~statham.schema.validation.format.format_checker`.
    """

    types = (str,)
    keywords = ("format",)
    message = "Must match format described by '{format}'."

    def _validate(self, value: Any):
        if not format_checker(self.params["format"], value):
            raise ValidationError



from typing import Any, ClassVar, Dict, Union

from statham.schema.elements.base import Element, UNBOUND_PROPERTY
from statham.schema.elements.meta import ObjectClassDict, ObjectMeta
from statham.schema.exceptions import ValidationError
from statham.schema.property import _Property
from statham.schema.constants import NotPassed


# TODO: Test and support limited recursive models.


class Object(metaclass=ObjectMeta):
    """Base model for JSON Schema ``"object"`` elements.

    ``"object"`` schemas are defined by declaring subclasses of
    :class:`Object`. Properties are declared as class attributes, and
    other keywords are set as class arguments.

    For example:

    .. code:: python

        from statham.schema.elements import Object, String
        from statham.schema.property import Property

        class Poll(Object, additionalProperties=False):

            questions: List[str] = Property(String(), required=True)

        poll = Poll({"questions": ["What's up?"]})

    """

    properties: ClassVar[Dict[str, _Property]]
    default: ClassVar[Any]
    additionalProperties: ClassVar[Union[Element, bool]]

    def __new__(
        cls, value: Any = NotPassed(), property_: _Property = UNBOUND_PROPERTY
    ):
        """Preprocess new instances.

        If value isn't passed, attempt to instantiate the default, but
        allow non-matching defaults.

        This is the equivalent of `Element.__call__`.
        """
        if isinstance(value, cls):
            return value
        if not isinstance(cls.default, NotPassed) and isinstance(
            value, NotPassed
        ):
            try:
                return cls(cls.default, property_)
            except (TypeError, ValidationError):
                return cls.default
        if isinstance(value, NotPassed):
            return value
        for validator in cls.validators:
            validator(value, property_)
        return object.__new__(cls)

    def __init__(
        self, value: Any = NotPassed(), _property: _Property = UNBOUND_PROPERTY
    ):
        """Initialise the object.

        The equivalent of `Element.__call__`, but on a class/instance.
        """
        if value is self:
            return
        if isinstance(value, NotPassed) and not isinstance(
            self.default, NotPassed
        ):
            value = self.default
        self._dict: Dict[str, Any] = {}
        for attr_name, attr_value in type(self).__properties__(value).items():
            if attr_name in type(self).properties:
                setattr(self, attr_name, attr_value)
            self._dict[attr_name] = attr_value

    def __repr__(self):
        attr_values = {
            attr: getattr(self, attr) for attr in type(self).properties
        }
        attr_repr = ", ".join(
            [
                f"{attr}={repr(value)}"
                for attr, value in attr_values.items()
                if not isinstance(value, NotPassed)
            ]
        )
        return f"{self.__class__.__name__}({attr_repr})"

    def __eq__(self, other):
        return (
            type(self) is type(other)
            # pylint: disable=protected-access
            and self._dict == other._dict
        )

    def __getitem__(self, key: str) -> Any:
        return self._dict[key]

    @staticmethod
    def inline(
        name: str, *, properties: Dict[str, Any] = None, **kwargs
    ) -> ObjectMeta:
        """Inline constructor for object schema elements.

        Useful for minor objects, at the cost of reduced type checking
        support.

        :param name: The name of the schema element.
        :param properties: Dictionary of properties accepted by the schema
            element.
        :param kwargs: Any accepted class args to `Object` subclasses.
        :return: A new subclass of `Object` with the appropriate validation
            rules.
        """
        properties = properties or {}
        object_properties = ObjectClassDict()
        for prop_name, prop in properties.items():
            object_properties[prop_name] = prop
        return ObjectMeta(name, (Object,), object_properties, **kwargs)



from typing import Any, List

from statham.schema.elements.base import Element
from statham.schema.constants import Maybe, NotPassed
from statham.schema.validation import InstanceOf


class Null(Element[None]):
    """JSON Schema ``"null"`` element."""

    def __init__(
        self,
        *,
        default: Maybe[None] = NotPassed(),
        const: Maybe[Any] = NotPassed(),
        enum: Maybe[List[Any]] = NotPassed(),
    ):
        self.default = default
        self.const = const
        self.enum = enum

    @property
    def annotation(self) -> str:
        return "None"

    @property
    def type_validator(self):
        return InstanceOf(type(None))



"""Serializers for in-memory Element trees.

Elements may be serialized to a JSON dictionary or a Python module.
"""
from statham.serializers.python import serialize_python
from statham.serializers.json import serialize_json



from typing import Set, Type, Union

from statham.schema.elements import Element, Object
from statham.schema.elements.meta import ObjectMeta
from statham.serializers.orderer import orderer, get_children


def serialize_python(*elements: Element) -> str:
    """Serialize schema elements to python declaration string.

    Captures declaration of the first Object elements, and any subsequent
    elements this depends on. Module imports and declaration order are
    dynamically inferred.

    :param elements: The :class:`~statham.schema.elements.Element` objects
        to serialize.
    :return: Python module contents as a string, declaring the element tree.
    """
    declarations = "\n\n".join(
        [object_model.python() for object_model in orderer(*elements)]
    )
    imports = _get_imports(declarations, *elements)
    return "\n\n\n".join(block for block in [imports, declarations] if block)


def _get_imports(declarations: str, *elements: Element) -> str:
    """Get import statements required by the elements."""
    imports = [
        _get_standard_imports(declarations),
        _get_statham_imports(declarations, *elements),
    ]
    return "\n\n".join(block for block in imports if block)


def _get_standard_imports(declaration: str) -> str:
    """Get group if imports from standard library.

    Currently only includes type annotations.
    """
    type_imports = ", ".join(
        [
            annotation
            for annotation in ("Any", "List", "Union")
            if annotation in declaration
        ]
    )
    if not type_imports:
        return ""
    return f"from typing import {type_imports}"


def _get_statham_imports(declaration: str, *elements: Element) -> str:
    """Construct imports from statham submodules."""
    statham_imports = []
    if "Maybe" in declaration:
        statham_imports.append("from statham.schema.constants import Maybe")
    element_imports = _get_element_imports(*elements)
    if element_imports:
        statham_imports.append(element_imports)
    if "Property" in declaration:
        statham_imports.append("from statham.schema.property import Property")
    return "\n".join(statham_imports)


def _get_element_imports(*elements: Element) -> str:
    """Get the import string for the elements in use."""
    prefix = "from statham.schema.elements import "
    max_length = 80
    import_names = [
        elem_type.__name__
        for elem_type in set.union(
            *(_get_single_element_imports(element) for element in elements)
        )
    ]
    if not import_names:
        return ""
    imports = ", ".join(sorted(import_names))
    if max_length < len(prefix) + len(imports):
        imports = "\n    ".join(["("] + imports.split(" ")) + ",\n)"
    return prefix + imports


def _get_single_element_imports(
    element: Element,
) -> Set[Union[Type[Element], ObjectMeta]]:
    """Extract the set of element types used by a given element."""
    get_type = (
        lambda elem: Object if isinstance(elem, ObjectMeta) else type(elem)
    )
    children = [element, *get_children(element)]
    return set(map(get_type, children))  # type: ignore



# pylint: disable=too-many-lines
"""Parsing tools to convert JSON Schema dictionaries to Element instances.

Some JSON Schema documents will be converted to an equivalent but structurally
differing representation. In particular, those that combine composition
keywords or use multiple types will be recomposed using ``"allOf"`` and
``"anyOf"`` respectively. See full docs for more
details.
"""
from collections import defaultdict
from functools import partial
import inspect
from itertools import chain
import operator as op
import re
import string
from typing import Any, Callable, DefaultDict, Dict, Iterable, List, Type, Union
import unicodedata

from statham.schema.constants import (
    COMPOSITION_KEYWORDS,
    NotPassed,
    UNSUPPORTED_SCHEMA_KEYWORDS,
)
from statham.schema.elements import (
    AllOf,
    AnyOf,
    Array,
    Boolean,
    CompositionElement,
    Integer,
    Not,
    Nothing,
    Null,
    Number,
    Object,
    OneOf,
    Element,
    String,
)
from statham.schema.elements.meta import (
    ObjectClassDict,
    ObjectMeta,
    RESERVED_PROPERTIES,
)
from statham.schema.exceptions import (
    FeatureNotImplementedError,
    SchemaParseError,
)
from statham.schema.helpers import expand, reraise, split_dict
from statham.schema.property import _Property


_TYPE_MAPPING = {
    "array": Array,
    "boolean": Boolean,
    "integer": Integer,
    "null": Null,
    "number": Number,
    "string": String,
}


class _ParseState:
    """Recusive state.

    Used to de-duplicate models which are traversed multiple times, and to
    rename distinct models with the same name.
    """

    def __init__(self):
        self.seen: DefaultDict[str, List[ObjectMeta]] = defaultdict(list)

    def dedupe(self, object_type: ObjectMeta):
        """Deduplicate a parsed model.

        If it has been seen before, then return the existing one. Otherwise
        ensure the model's name is distinct from other models and keep store
        it.
        """
        name = object_type.__name__
        for existing in self.seen[name]:
            if object_type == existing:
                return existing
        count = len(self.seen[name])
        if count:
            object_type.__name__ = name + f"_{count}"
        self.seen[name].append(object_type)
        return object_type


def parse(schema: Dict[str, Any]) -> List[Element]:
    """Parse a JSON Schema document to Element format.

    Assumes references are already resolved, and that any ``"object"`` schemas
    or sub-schemas contain either a ``"title"`` annotation or an
    ``"_x_autotitle"`` annotation. See
    `json-ref-dict <https://pypi.org/project/json-ref-dict/0.6.0/>`_ for
    reference resolution and annotation tools.

    :return: A list of schema  elements, starting with the top level element,
        followed by each element in the top-level schema ``"definitions"``.
    """
    state = _ParseState()
    return [parse_element(schema, state)] + [
        parse_element(definition, state)
        for definition in schema.get("definitions", {}).values()
        if isinstance(definition, (dict, bool, Element))
    ]


@reraise(
    RecursionError,
    FeatureNotImplementedError,
    "Could not parse cyclical dependencies of this schema.",
)
def parse_element(
    schema: Union[bool, Dict[str, Any]], state: _ParseState = None
) -> Element:
    """Parse a single JSON Schema element to an Element object.

    Called by :func:`parse` when parsing entire documents.

    >>> parse_element({"type": "string", "minLength": 3})
    String(minLength=3)

    :raises: :exc:`~statham.schema.exceptions.FeatureNotImplementedError` if
        recursive cycles are detected.
    :raises: :exc:`statham.schema.exceptions.SchemaParseError` if problems are
        found in the provided schema.
    :return: A single :class:`~statham.schema.elements.Element` object
        equivalent to the schema described by :paramref:`parse_element.schema`.
    """
    if isinstance(schema, bool):
        return Element() if schema else Nothing()
    state = state or _ParseState()
    if isinstance(schema, Element):
        return schema
    if set(schema) & UNSUPPORTED_SCHEMA_KEYWORDS:
        raise FeatureNotImplementedError.unsupported_keywords(
            set(schema) & UNSUPPORTED_SCHEMA_KEYWORDS
        )
    for literal_key in ("default", "const", "enum"):
        if literal_key in schema:
            schema[literal_key] = _parse_literal(schema[literal_key])
    for keyword, parser in (
        ("properties", _parse_properties),
        ("items", _parse_items),
        ("patternProperties", _parse_pattern_properties),
        ("propertyNames", _parse_property_names),
        ("contains", _parse_contains),
        ("dependencies", _parse_dependencies),
    ):
        if keyword in schema:
            schema[keyword] = parser(schema, state)  # type: ignore
    schema["additionalProperties"] = _parse_additional_properties(schema, state)
    schema["additionalItems"] = _parse_additional_items(schema, state)
    if set(COMPOSITION_KEYWORDS) & set(schema):
        return _parse_composition(schema, state)
    if "type" not in schema:
        return Element(**_keyword_filter(Element)(schema))
    return _parse_typed(schema["type"], schema, state)


def _parse_literal(literal: Any) -> Any:
    """Parse literal values from schema.

    Keywords like `const`, `enum` and `default` refer to non-schema values.
    Annotations should be removed to prevent side effects.
    """
    if not isinstance(literal, (dict, list)):
        return literal
    if isinstance(literal, list):
        return [_parse_literal(val) for val in literal]
    return {
        key: _parse_literal(val)
        for key, val in literal.items()
        if key != "_x_autotitle"
    }


def _parse_contains(
    schema: Dict[str, Any], state: _ParseState = None
) -> Element:
    """Parse schema contains keyword."""
    state = state or _ParseState()
    return parse_element(schema["contains"], state)


def _parse_composition(
    schema: Dict[str, Any], state: _ParseState = None
) -> Element:
    """Parse a schema with composition keywords.

    Handles multiple composition keywords by wrapping them in an AllOf
    element. Similarly, non-keyword elements are parsed as usual and
    included in an allOf element.

    For example:
    ```python
    schema = {
        "type": "integer",
        "oneOf": [{"minimum": 3}, {"maximum": 5}],
        "anyOf": [{"multipleOf": 2}, {"multipleOf": 3}],
    }
    _parse_composition(schema) == AllOf(
        Integer(),
        OneOf(Element(minimum=3), Element(maximum=5)),
        AnyOf(Element(multipleOf=2), Element(multipleOf=3)),
    )
    ```
    """
    state = state or _ParseState()
    composition, other = split_dict(set(COMPOSITION_KEYWORDS) | {"default"})(
        schema
    )
    base_element = parse_element(other, state)
    for key in set(COMPOSITION_KEYWORDS) - {"not"}:
        composition[key] = [
            parse_element(sub_schema, state)
            for sub_schema in composition.get(key, [])
        ]
    all_of = [base_element] + composition["allOf"]
    all_of.append(_compose_elements(OneOf, composition["oneOf"]))
    all_of.append(_compose_elements(AnyOf, composition["anyOf"]))
    if "not" in composition:
        all_of.append(Not(parse_element(schema["not"], state)))
    element = _compose_elements(
        AllOf, filter(partial(op.ne, Element()), all_of)
    )
    default = schema.get("default", NotPassed())
    if isinstance(element, ObjectMeta):
        return AllOf(element, default=default)
    element.default = default or element.default
    return element


def _parse_typed(
    type_value: Any, schema: Dict[str, Any], state: _ParseState = None
) -> Element:
    """Parse a typed schema with no composition keywords."""
    state = state or _ParseState()
    if not isinstance(type_value, (str, list)):
        raise SchemaParseError.invalid_type(type_value)
    if isinstance(type_value, list):
        return _parse_multi_typed(type_value, schema, state)
    if schema["type"] == "object":
        return _parse_object(schema, state)
    if schema["type"] == "array":
        return _parse_array(schema, state)
    element_type = _TYPE_MAPPING[type_value]
    sub_schema = _keyword_filter(element_type)(schema)
    return element_type(**sub_schema)


def _parse_multi_typed(
    type_list: List[str], schema: Dict[str, Any], state: _ParseState = None
) -> CompositionElement:
    """Parse a schema with multiple type values.

    Converts schema to an equivalent representation using "anyOf". For example:
    ```python
    {"type": ["string", "integer"]}
    ```
    becomes
    ```
    {"anyOf": [{"type": "string"}, {"type": "integer"}]}
    ```
    """
    state = state or _ParseState()
    default = schema.get("default", NotPassed())
    schema = {key: val for key, val in schema.items() if key != "default"}
    if len(type_list) == 1:
        return parse_element({**schema, "type": type_list[0]}, state)
    return AnyOf(
        *(
            parse_element({**schema, "type": type_value}, state)
            for type_value in type_list
        ),
        default=default,
    )


def _parse_object(
    schema: Dict[str, Any], state: _ParseState = None
) -> ObjectMeta:
    """Parse an object schema element to an `Object` subclass.

    The name of the generated class is derived from the following keys
    in precedence order:
    - "title"
    - "_x_autotitle"

    Also requires that "properties" and "additionalProperties" values have
    already been parsed.

    :raises SchemaParseError: if keys exist from which to derive the class
        title.
    :return: The ``Object`` model equivalent to the schema.
    """
    state = state or _ParseState()
    title = schema.get("title", schema.get("_x_autotitle"))
    if not title:
        raise SchemaParseError.missing_title(schema)
    title = _title_format(title)
    properties = schema.get("properties", {})
    properties.update(
        {
            _parse_attribute_name(key): _Property(
                Element(), required=True, source=key
            )
            for key in schema.get("required", [])
            if _parse_attribute_name(key) not in properties
        }
    )
    class_dict = ObjectClassDict()
    for key, value in properties.items():
        class_dict[key] = value
    cls_args = dict(additionalProperties=schema["additionalProperties"])
    for key in [
        "patternProperties",
        "minProperties",
        "maxProperties",
        "propertyNames",
        "dependencies",
        "const",
        "enum",
        "default",
    ]:
        if key in schema:
            cls_args[key] = schema[key]
    object_type = ObjectMeta(title, (Object,), class_dict, **cls_args)
    return state.dedupe(object_type)


def _parse_properties(
    schema: Dict[str, Any], state: _ParseState = None
) -> Dict[str, _Property]:
    """Parse properties from a schema element."""
    state = state or _ParseState()
    required = set(schema.get("required", []))
    properties = schema.get("properties", {})
    return {
        **{
            _parse_attribute_name(key): _Property(
                parse_element(value, state),
                required=key in required,
                source=key,
            )
            for key, value in properties.items()
            # Ignore malformed values.
            if isinstance(value, (dict, bool))
        },
        **{
            _parse_attribute_name(key): prop
            for key, prop in properties.items()
            if isinstance(prop, _Property)
        },
    }


def _parse_attribute_name(name: str) -> str:
    """Convert attibute name to valid python attribute.

    Attempts to replace special characters with their unicode names.
    """

    def _char_map(idx: int, char: str) -> str:
        if char.isalnum() or char in ("_", "-", " "):
            return char
        if char in string.whitespace:
            return "_"
        label = unicodedata.name(char, "unknown").lower()
        if idx != 0 and name[idx - 1] != "_":
            label = "_" + label
        if idx != len(name) - 1 and name[idx + 1] != "_":
            label = label + "_"
        return label

    chars = map(expand(_char_map), enumerate(name))
    name = "".join(chars).replace(" ", "_").replace("-", "_")
    if not name:
        return "blank"
    first_chars = set(string.ascii_letters) | {"_"}
    if name[0] not in first_chars:
        name = f"_{name}"
    if name in RESERVED_PROPERTIES:
        name = f"{name}_"
    return name


def _parse_pattern_properties(
    schema: Dict[str, Any], state: _ParseState = None
) -> Dict[str, Element]:
    """Parse schema patternProperties keyword."""
    state = state or _ParseState()
    return {
        **{
            key: parse_element(value, state)
            for key, value in schema["patternProperties"].items()
            if isinstance(value, (dict, bool))
        },
        **{
            key: value
            for key, value in schema["patternProperties"].items()
            if isinstance(value, Element)
        },
    }


def _parse_property_names(
    schema: Dict[str, Any], state: _ParseState = None
) -> Element:
    """Parse schema propertyNames keyword."""
    state = state or _ParseState()
    return parse_element(schema["propertyNames"], state)


def _parse_additional(
    key: str, schema: Dict[str, Any], state: _ParseState = None
) -> Union[Element, bool]:
    """Parse additional items or properties.

    Booleans are retained for these values, as they are more semantically
    meaningful than in general schemas.
    """
    state = state or _ParseState()
    additional = schema.get(key, True)
    if isinstance(additional, bool):
        return additional
    return parse_element(additional, state)


def _parse_additional_properties(
    schema: Dict[str, Any], state: _ParseState = None
) -> Union[Element, bool]:
    """Parse additionalProperties from a schema element.

    If key is not present, defaults to `True`.
    """
    return _parse_additional("additionalProperties", schema, state)


def _parse_additional_items(
    schema: Dict[str, Any], state: _ParseState = None
) -> Union[Element, bool]:
    """Parse additionalProperties from a schema element.

    If key is not present, defaults to `True`.
    """
    return _parse_additional("additionalItems", schema, state)


def _parse_array(schema: Dict[str, Any], state: _ParseState = None) -> Array:
    """Parse an array schema element."""
    state = state or _ParseState()
    items = schema.get("items", Element())
    return Array(**{**_keyword_filter(Array)(schema), "items": items})


def _parse_items(
    schema: Dict[str, Any], state: _ParseState = None
) -> Union[Element, List[Element]]:
    """Parse array items keyword to a schema Element.

    If not present, defaults to `Element()`.
    """
    state = state or _ParseState()
    items = schema.get("items", {})
    if isinstance(items, list):
        return [parse_element(item, state) for item in items]
    return parse_element(items, state)


def _parse_dependencies(
    schema: Dict[str, Any], state: _ParseState = None
) -> Dict[str, Union[List[str], Element]]:
    """Parse dependencies keyword from schema."""
    state = state or _ParseState()
    return {
        **{
            key: value
            for key, value in schema["dependencies"].items()
            if isinstance(value, (list, Element))
        },
        **{
            key: parse_element(value, state)
            for key, value in schema["dependencies"].items()
            if isinstance(value, (dict, bool))
        },
    }


def _compose_elements(
    element_type: Type[CompositionElement], elements: Iterable[Element]
) -> Element:
    """Create a composition element from a type and list of component elements.

    Filters out trivial elements, and simplifies compositions with only one
    composed element.
    """
    elements = list(elements)
    if not elements:
        return Element()
    if len(elements) == 1:
        return elements[0]
    return element_type(*elements)


def _keyword_filter(type_: Type) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    """Create a filter to pull out only relevant keywords for a given type."""
    params = inspect.signature(type_.__init__).parameters.values()
    args = {param.name for param in params}

    def _filter(schema: Dict[str, Any]) -> Dict[str, Any]:
        return {key: value for key, value in schema.items() if key in args}

    return _filter


def _title_format(name: str) -> str:
    """Convert titles in schemas to class names."""
    words = list(filter(None, re.split("[^a-zA-Z0-9]", name)))
    segments = chain.from_iterable(
        [
            re.findall("[A-Z][^A-Z]*", word[0].upper() + word[1:])
            for word in words
        ]
    )
    return "".join(segment.title() for segment in segments)



from typing import Any, cast, Optional

from statham.schema.exceptions import ValidationError
from statham.schema.helpers import remove_duplicates
from statham.schema.validation.base import Validator, replace_bool


class MinItems(Validator):
    """Validate that arrays have a minimum number of items."""

    types = (list,)
    keywords = ("minItems",)
    message = "Must contain at least {minItems} items."

    def _validate(self, value: Any):
        if len(value) < self.params["minItems"]:
            raise ValidationError


class MaxItems(Validator):
    """Validate that arrays have a maximum number of items."""

    types = (list,)
    keywords = ("maxItems",)
    message = "Must contain fewer than {maxItems} items."

    def _validate(self, value: Any):
        if len(value) > self.params["maxItems"]:
            raise ValidationError


class AdditionalItems(Validator):
    """Validate array items not covered by the ``"items"`` keyword.

    Only relevant when using "tuple" style ``"items"``.
    """

    types = (list,)
    keywords = ("items", "additionalItems")
    message = "Must not contain additional items. Accepts: {items}"

    def _validate(self, value: Any):
        if not isinstance(self.params["items"], list):
            return
        if len(value) <= len(self.params["items"]):
            return
        if self.params["additionalItems"]:
            return
        raise ValidationError


class UniqueItems(Validator):
    """Validate that array items are unique."""

    types = (list,)
    keywords = ("uniqueItems",)
    message = "Must not contain duplicates."

    @classmethod
    def from_element(cls, element) -> Optional["UniqueItems"]:
        validator: Optional[UniqueItems] = cast(
            Optional[UniqueItems], super().from_element(element)
        )
        if validator and validator.params["uniqueItems"] is False:
            return None
        return validator

    def _validate(self, value: Any):
        # Once again, Cpython's 1 in [True] nightmare.
        aliased_value = list(map(replace_bool, value))
        length = len(aliased_value)
        try:
            # Try the hashable approach
            if len(set(aliased_value)) == length:
                return
        except TypeError:
            # Long version
            if len(remove_duplicates(aliased_value)) == length:
                return
        raise ValidationError


class Contains(Validator):
    """Validate that at least one array item matches a schema."""

    types = (list,)
    keywords = ("contains",)
    message = "Must contain one element matching {contains}."

    def _validate(self, value: Any):
        for sub_value in value:
            try:
                _ = self.params["contains"](sub_value)
                return
            except (TypeError, ValidationError):
                continue
        raise ValidationError



from typing import Any
from statham.schema.exceptions import ValidationError
from statham.schema.validation.base import Validator


class Minimum(Validator):
    """Validate that numeric values conform to a minimum."""

    types = (int, float)
    keywords = ("minimum",)
    message = "Must be greater than or equal to {minimum}."

    def _validate(self, value: Any):
        if value < self.params["minimum"]:
            raise ValidationError


class Maximum(Validator):
    """Validate that numeric values conform to a maximum."""

    types = (int, float)
    keywords = ("maximum",)
    message = "Must be less than or equal to {maximum}."

    def _validate(self, value: Any):
        if value > self.params["maximum"]:
            raise ValidationError


class ExclusiveMinimum(Validator):
    """Validate that numeric values conform to an exclusive minimum."""

    types = (int, float)
    keywords = ("exclusiveMinimum",)
    message = "Must be strictly greater than {exclusiveMinimum}."

    def _validate(self, value: Any):
        if value <= self.params["exclusiveMinimum"]:
            raise ValidationError


class ExclusiveMaximum(Validator):
    """Validate that numeric values conform to an exclusive maximum."""

    types = (int, float)
    keywords = ("exclusiveMaximum",)
    message = "Must be strictly less than {exclusiveMaximum}."

    def _validate(self, value: Any):
        if value >= self.params["exclusiveMaximum"]:
            raise ValidationError


class MultipleOf(Validator):
    """Validate that numeric values are a multiple of a given number."""

    types = (int, float)
    keywords = ("multipleOf",)
    message = "Must be a multiple of {multipleOf}."

    def _validate(self, value: Any):
        multiple_of = self.params["multipleOf"]
        if isinstance(multiple_of, float):
            quotient = value / multiple_of
            if int(quotient) != quotient:
                raise ValidationError
            return
        if value % multiple_of:
            raise ValidationError



import re
from typing import Any, Dict, Generic, Iterator, TypeVar

from statham.schema.constants import NotPassed
from statham.schema.elements.base import Element, Nothing
from statham.schema.elements.composition import AllOf
from statham.schema.property import _Property as Property


class Properties:
    """Interface for retrieving relevant schemas given a property name.

    Used internally by :class:`~statham.schema.elements.Element`.
    """

    def __init__(self, element, props, pattern=None, additional=True):
        self.element = element
        self.props = props or {}
        self.pattern = PatternDict(pattern or {})
        for name, prop in self.props.items():
            prop.bind(name=name, parent=self.element)
        if isinstance(additional, bool):
            self.additional = {True: Element(), False: Nothing()}[additional]
        else:
            self.additional = additional

    def property(self, element, name):
        prop = Property(element)
        prop.bind(name=name, parent=self.element)
        return prop

    def __repr__(self):
        props = [repr(self.props)]
        if self.pattern:
            props.append(f"patternProperties={self.pattern}")
        if self.additional == Nothing():
            props.append("additionalProperties=False")
        elif self.additional != Element():
            props.append(f"additionalProperties={self.additional}")
        return f"{type(self).__name__}({', '.join(props)})"

    def __getitem__(self, key):
        prop = {prop.source: prop for prop in self.props.values()}.get(
            key, None
        )
        pattern_elems = list(self.pattern.getall(key))
        if not (prop or pattern_elems):
            return self.property(self.additional, key)
        if not prop:
            if len(pattern_elems) == 1:
                return self.property(pattern_elems[0], key)
            return self.property(AllOf(*pattern_elems), key)
        if not pattern_elems:
            return prop
        composite = Property(
            AllOf(prop.element, *pattern_elems),
            source=prop.source,
            required=prop.required,
        )
        composite.bind(name=prop.name, parent=prop.parent)
        return composite

    def __contains__(self, key):
        return bool(self[key].element != Nothing())

    def __iter__(self):
        return iter(self.props)

    def __call__(self, value):
        value = {
            **{prop.name: NotPassed() for prop in self.props.values()},
            **value,
        }
        return {
            self[key].name or key: self[key](sub_value)
            for key, sub_value in value.items()
        }


T = TypeVar("T")


class PatternDict(Dict[str, T], Generic[T]):
    """Dict for simplifying PatternProperties lookups."""

    def __getitem__(self, key: str) -> T:
        if not isinstance(key, str):
            raise KeyError
        try:
            return next(self.getall(key))
        except StopIteration:
            raise KeyError

    def getall(self, key: str) -> Iterator[T]:
        for pattern, value in self.items():  # pylint: disable=no-member
            if re.search(pattern, key):
                yield value

    def __contains__(self, key: Any) -> bool:
        try:
            _ = self[key]
            return True
        except KeyError:
            return False



import inspect
import keyword
from typing import Any, cast, Dict, List, Tuple, Type, Union

from statham.schema.constants import Maybe, NotPassed
from statham.schema.elements.base import Element
from statham.schema.property import _Property
from statham.schema.exceptions import SchemaDefinitionError
from statham.schema.validation import (
    AdditionalProperties,
    Const,
    Dependencies,
    Enum,
    InstanceOf,
    MaxProperties,
    MinProperties,
    PropertyNames,
    Required,
    Validator,
)


RESERVED_PROPERTIES = dir(object) + list(keyword.kwlist) + ["_dict"]


class ObjectClassDict(dict):
    """Overriden class dictionary for the metaclass of Object.

    Collects schema properties and default value if present.
    """

    properties: Dict

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.properties = {}

    def __setitem__(self, key, value):
        if key in RESERVED_PROPERTIES and isinstance(value, _Property):
            raise SchemaDefinitionError.reserved_attribute(key)
        if isinstance(value, _Property):
            return self.properties.__setitem__(key, value)
        return super().__setitem__(key, value)


class ObjectMeta(type, Element):
    """Metaclass to allow declaring Object schemas as classes.

    Collects default value and properties defined as class variables,
    and binds information to those properties.
    """

    additionalProperties: Union[Element, bool]
    patternProperties: Maybe[Dict[str, Element]]
    minProperties: Maybe[int]
    maxProperties: Maybe[int]
    propertyNames: Maybe[Element]
    const: Maybe[Any]
    enum: Maybe[List[Any]]
    dependencies: Maybe[Dict[str, Union[List[str], Element]]]

    @staticmethod
    def __subclasses__():
        # This is overriden to prevent errors.
        # TODO: Is there a more elegant way to achieve this? Perhaps
        #   __init_subclass__ should error to prevent this from being
        #   wrong.
        return []

    @classmethod
    def __prepare__(mcs, _name, _bases, **_kwargs):
        return ObjectClassDict()

    # pylint: disable=too-many-locals
    def __new__(
        mcs,
        name: str,
        bases: Tuple[Type],
        classdict: ObjectClassDict,
        *,
        default: Maybe[Any] = NotPassed(),
        const: Maybe[Any] = NotPassed(),
        enum: Maybe[List[Any]] = NotPassed(),
        required: Maybe[List[str]] = NotPassed(),
        minProperties: Maybe[int] = NotPassed(),
        maxProperties: Maybe[int] = NotPassed(),
        patternProperties: Maybe[Dict[str, Element]] = NotPassed(),
        additionalProperties: Maybe[Union[Element, bool]] = NotPassed(),
        propertyNames: Maybe[Element] = NotPassed(),
        dependencies: Maybe[Dict[str, Union[List[str], Element]]] = NotPassed(),
    ):
        cls: ObjectMeta = cast(
            ObjectMeta, type.__new__(mcs, name, bases, dict(classdict))
        )
        previous = lambda attr, default: getattr(cls, attr, default)
        get_value = (
            lambda value, attr: value
            if not isinstance(value, NotPassed)
            else previous(attr, NotPassed())
        )
        cls.default = get_value(default, "default")
        cls.const = get_value(const, "const")
        cls.enum = get_value(enum, "enum")
        cls.required = get_value(required, "required")
        # https://github.com/python/mypy/issues/3004
        cls.properties = {  # type: ignore
            **{
                attr: prop.clone()
                for attr, prop in previous("properties", {}).items()
            },
            **classdict.properties,
        }
        cls.minProperties = get_value(minProperties, "minProperties")
        cls.maxProperties = get_value(maxProperties, "maxProperties")
        cls.patternProperties = get_value(
            patternProperties, "patternProperties"
        )
        cls.additionalProperties = (
            additionalProperties
            if not isinstance(additionalProperties, NotPassed)
            else previous("additionalProperties", True)
        )
        cls.propertyNames = get_value(propertyNames, "propertyNames")
        cls.dependencies = get_value(dependencies, "dependencies")
        return cls

    def __hash__(cls):
        return hash(tuple([cls.__name__] + list(cls.properties or [])))

    @property
    def annotation(cls) -> str:
        return cls.__name__

    def __repr__(cls):
        return cls.__name__

    @property
    def type_validator(cls) -> Validator:
        return InstanceOf(dict, cls)

    @property
    def validators(cls) -> List[Validator]:
        possible_validators = [
            cls.type_validator,
            Required.from_element(cls),
            AdditionalProperties(cls.__properties__),
            MinProperties.from_element(cls),
            MaxProperties.from_element(cls),
            PropertyNames.from_element(cls),
            Const.from_element(cls),
            Enum.from_element(cls),
            Dependencies.from_element(cls),
        ]
        return [validator for validator in possible_validators if validator]

    def python(cls) -> str:
        super_cls = next(iter(cls.mro()[1:]))
        cls_args = [super_cls.__name__]
        parameters = list(
            inspect.signature(type(cls).__new__).parameters.values()
        )
        for param in parameters:
            if param.kind != param.KEYWORD_ONLY:
                continue
            value = getattr(cls, param.name, NotPassed())
            if value == param.default or (
                param.name == "additionalProperties" and value is True
            ):
                continue
            cls_args.append(f"{param.name}={repr(value)}")
        class_def = f"""class {repr(cls)}({', '.join(cls_args)}):
"""
        if not cls.properties:
            class_def = (
                class_def
                + """
    pass
"""
            )
        # False positive
        # pylint: disable=no-member
        for property_ in cast(
            Dict[str, _Property], cls.properties or {}
        ).values():
            class_def = (
                class_def
                + f"""
    {property_.python()}
"""
            )
        return class_def



"""This module contains implementation of JSON Schema validation keywords.

Each keyword is implemented as a subclass of
:class:`~statham.schema.validation.base.Validator`, and is instantiated with
the relevant keywords. :class:`~statham.schema.validation.base.Validator`
instances may then be called to validate values.

:class:`~statham.schema.elements.Element` automatically detects and instantiates
its relevant validators, but validators may be used directly:

.. code:: python

    from statham.schema.validation import Minimum

    validator = Minimum(minimum=3)
    validator(5, None)  # OK
    validator(2, None)  # ValidationError

"""
from typing import Iterator, Type

from statham.schema.validation.array import (
    AdditionalItems,
    Contains,
    MinItems,
    MaxItems,
    UniqueItems,
)
from statham.schema.validation.base import (
    Const,
    Enum,
    InstanceOf,
    NoMatch,
    Validator,
)
from statham.schema.validation.format import format_checker
from statham.schema.validation.numeric import (
    Minimum,
    Maximum,
    ExclusiveMinimum,
    ExclusiveMaximum,
    MultipleOf,
)
from statham.schema.validation.object import (
    AdditionalProperties,
    Dependencies,
    MaxProperties,
    MinProperties,
    PropertyNames,
    Required,
)
from statham.schema.validation.string import (
    MinLength,
    MaxLength,
    Format,
    Pattern,
)


def _all_subclasses(klass: Type):
    """Get all explicit and implicit subclasses of a type."""
    return set(klass.__subclasses__()).union(
        [s for c in klass.__subclasses__() for s in _all_subclasses(c)]
    )


def get_validators(element) -> Iterator[Validator]:
    """Iterate all applicable validators for an Element.

    Validators identify whether they are applicable for an element via the
    `from_element` class method. In general, this checks whether its
    parameters are present on the element with correct values.
    """
    for validator_type in _all_subclasses(Validator):
        if validator_type in (InstanceOf, NoMatch):
            continue
        validator = validator_type.from_element(element)
        if validator:
            yield validator



from statham.schema.elements.array import Array
from statham.schema.elements.base import Element, Nothing
from statham.schema.elements.boolean import Boolean
from statham.schema.elements.composition import (
    AllOf,
    AnyOf,
    CompositionElement,
    Not,
    OneOf,
)
from statham.schema.elements.null import Null
from statham.schema.elements.numeric import Integer, Number
from statham.schema.elements.object import Object
from statham.schema.elements.string import String



from typing import Callable, Tuple

from json_ref_dict import URI

from statham.schema.constants import COMPOSITION_KEYWORDS


def _get_title_from_reference(reference: str) -> str:
    """Convert JSON Schema references to title fields.

    If the reference has a pointer, use the final segment, otherwise
    use the final segment of the base uri stripping any content type
    extension.

    :param reference: The JSONPointer reference.
    :return: An autogenerated title for the reference target.
    """
    uri = URI.from_string(reference)
    if not uri.pointer.strip("/"):
        return uri.uri_name.split(".")[0]
    title = uri.pointer.split("/")[-1]
    if title == "items":
        return _get_title_from_reference(repr(uri.back())) + "Item"
    if title.isdigit():
        # Handle anonymous schemas in an array.
        # E.G `{"anyOf": [{"type": "object"}]}`
        return _get_title_from_reference(repr(uri.back())) + title
    if title in COMPOSITION_KEYWORDS:
        return _get_title_from_reference(repr(uri.back()))
    return title


def title_labeller() -> Callable[[str], Tuple[str, str]]:
    """Create a title labeller, enumerating repeated titles.

    Used to assign meaningful names to schemas which have no specified
    title.
    """

    def _get_title(reference: str) -> Tuple[str, str]:
        name = _get_title_from_reference(reference)
        return "_x_autotitle", name

    return _get_title



from typing import Callable, Dict
from uuid import UUID
import warnings

from dateutil.parser import parse as parse_datetime, ParserError  # type: ignore


class _FormatString:
    """Extendable format string register.

    # TODO: Built-in formats.
    """

    def __init__(self, name: str):
        self._callable_register: Dict[str, Callable[[str], bool]] = {}
        self.__name__: str = name

    def register(self, format_string: str) -> Callable:
        def _register_callable(is_format: Callable[[str], bool]):
            self._callable_register[format_string] = is_format

        return _register_callable

    def __call__(self, format_string: str, value: str) -> bool:
        if format_string not in self._callable_register:
            full_name = f"{self.__module__}.{self.__name__}"
            warnings.warn(
                (
                    f"""No validator found for format string {format_string}.
To register new formats, please register a checker with
{full_name} as follows:
```
@{self.__name__}({format_string})
def is_{format_string.replace('-', '_')}(value) -> bool:
    ...
```
"""
                ),
                RuntimeWarning,
            )
            return True
        return self._callable_register[format_string](value)


format_checker: _FormatString = _FormatString("format_checker")
"""Extensible implementation of the ``"format"`` keyword.

Validators for new formats may be registered as follows:

.. code:: python

    @format_checker.register("my_format")
    def _validate_my_format(value: str) -> bool:
        # Return True if `value` matches `my_format`.
        ...
"""


@format_checker.register("uuid")
def _is_uuid(value: str) -> bool:
    try:
        UUID(value)
    except (ValueError, TypeError):
        return False
    return True


@format_checker.register("date-time")
def _is_date_time(value: str) -> bool:
    try:
        parse_datetime(value)
    except (ParserError, TypeError):
        return False
    return True



from functools import wraps
import inspect
from typing import (
    Any,
    Callable,
    Container,
    Dict,
    Iterable,
    List,
    Tuple,
    Type,
    TypeVar,
    Union,
)


class Args:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def apply(self, function):
        return function(*self.args, **self.kwargs)

    def __repr__(self):
        arg_string = ", ".join([repr(arg) for arg in self.args])
        kwarg_string = ", ".join(
            [f"{key}={repr(value)}" for key, value in self.kwargs.items()]
        )
        return "(" + ", ".join(filter(None, [arg_string, kwarg_string])) + ")"


def custom_repr_args(self, **overrides: Any) -> Args:
    args: List[Any] = []
    kwargs: Dict[str, Any] = {}
    parameters = list(
        inspect.signature(type(self).__init__).parameters.values()
    )[1:]
    for param in parameters:
        value = overrides.get(param.name, getattr(self, param.name, None))
        if value == param.default:
            continue
        if param.kind == param.VAR_POSITIONAL:
            args.extend(value or [])
        elif param.kind == param.KEYWORD_ONLY:
            kwargs[param.name] = value
        else:
            args.append(value)
    return Args(*args, **kwargs)


def custom_repr(self, **overrides: Any) -> str:
    """Dynamically construct the repr to match value instantiation.

    Shows the class name and attribute values, where they differ from
    defaults.

    :param overrides: Overwrites of instance attribute values for the
        repr.
    :return: String representation of :paramref:`custom_repr.self`.
    """
    return f"{type(self).__name__}{repr(custom_repr_args(self, **overrides))}"


ExceptionTypes = Union[Type[Exception], Tuple[Type[Exception], ...]]


def reraise(catch: ExceptionTypes, throw: Type[Exception], message: str):
    """Decorator factory for re-raising exceptions of a raised in a function."""

    def _decorator(function):
        @wraps(function)
        def _wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except catch as exc:
                raise throw(message) from exc

        return _wrapper

    return _decorator


T = TypeVar("T")


def split_dict(
    keys: Container[T],
) -> Callable[[Dict[T, Any]], Tuple[Dict[T, Any], Dict[T, Any]]]:
    """Split a dictionary on matching and non-matching keys."""

    def _split(dictionary: Dict[T, Any]) -> Tuple[Dict[T, Any], Dict[T, Any]]:
        match = {key: value for key, value in dictionary.items() if key in keys}
        complement = {
            key: value for key, value in dictionary.items() if key not in keys
        }
        return match, complement

    return _split


def expand(function):
    """Useful for functional operations like map."""
    return lambda args: function(*args)


SequenceItem = TypeVar("SequenceItem")


def remove_duplicates(seq: Iterable[SequenceItem]) -> List[SequenceItem]:
    """Remove duplicates whilst preserving order."""
    seen: List[SequenceItem] = []
    seen_add = seen.append
    return [x for x in seq if not (x in seen or seen_add(x))]



from argparse import ArgumentParser, RawTextHelpFormatter
from contextlib import contextmanager
from logging import getLogger, INFO
from os import path
from typing import Iterator, TextIO, Tuple
from sys import argv, stdout

from json_ref_dict import materialize, RefDict

from statham.schema.parser import parse
from statham.serializers import serialize_python
from statham.titles import title_labeller


LOGGER = getLogger(__name__)
LOGGER.setLevel(INFO)


def parse_input_arg(input_arg: str) -> str:
    """Parse input URI as a valid JSON Schema ref.

    This tool accepts bare base URIs, without the JSON Pointer,
    so these should be converted to a root pointer.
    """
    if "#" not in input_arg:
        return input_arg + "#/"
    return input_arg


@contextmanager
def parse_args(args) -> Iterator[Tuple[str, TextIO]]:
    """Parse arguments, abstracting IO in a context manager."""

    parser = ArgumentParser(
        description="Generate statham models from JSON Schema files.",
        formatter_class=RawTextHelpFormatter,
        add_help=False,
    )
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "--input",
        type=str,
        required=True,
        help="""Specify the path to the JSON Schema to be generated.

If the target schema is not at the root of a document, specify the
JSON Pointer in the same format as a JSON Schema `$ref`, e.g.
`--input path/to/document.json#/definitions/schema`

""",
    )
    optional = parser.add_argument_group("Optional arguments")
    optional.add_argument(
        "--output",
        type=str,
        default=None,
        help="""Output directory or file in which to write the output.

If the provided path is a directory, the command will derive the name
from the input argument. If not passed, the command will write to
stdout.

""",
    )
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        help="Display this help message and exit.",
    )
    parsed = parser.parse_args(args)
    input_arg: str = parse_input_arg(parsed.input)
    if parsed.output:
        if path.isdir(parsed.output):
            filename = ".".join(path.basename(parsed.input).split(".")[:-1])
            output_path = path.join(parsed.output, ".".join([filename, "py"]))
        else:
            output_path = parsed.output
        with open(output_path, "w", encoding="utf8") as file:
            yield input_arg, file
        return
    yield input_arg, stdout
    return


def main(input_uri: str) -> str:
    """Get a schema from a URI, and then return the generated python module.

    :param input_uri: URI of the target schema. This must follow the conventions
        of a JSON Schema ``"$ref"`` attribute.
    :return: Python module contents for generated models, as a string.
    """
    schema = materialize(
        RefDict.from_uri(input_uri), context_labeller=title_labeller()
    )
    return serialize_python(*parse(schema))


def entry_point():
    """Entry point for command.

    Parse arguments, read from input and write to output.
    """
    with parse_args(argv[1:]) as (uri, output):  # pragma: no cover
        output.write(main(uri))  # pragma: no cover


if __name__ == "__main__":  # pragma: no cover
    entry_point()  # pragma: no cover



from typing import Any, List, TypeVar, Union

from statham.schema.elements.base import Element
from statham.schema.constants import Maybe, NotPassed
from statham.schema.validation import InstanceOf


T = TypeVar("T", int, float)

Numeric = Union[int, float]


# pylint: disable=too-many-instance-attributes
class NumericElement(Element[T]):
    """JSON Schema base numeric element."""

    def __init__(
        self,
        *,
        default: Maybe[T] = NotPassed(),
        const: Maybe[Any] = NotPassed(),
        enum: Maybe[List[Any]] = NotPassed(),
        minimum: Maybe[Numeric] = NotPassed(),
        maximum: Maybe[Numeric] = NotPassed(),
        exclusiveMinimum: Maybe[Numeric] = NotPassed(),
        exclusiveMaximum: Maybe[Numeric] = NotPassed(),
        multipleOf: Maybe[Numeric] = NotPassed(),
    ):
        self.default = default
        self.const = const
        self.enum = enum
        self.minimum = minimum
        self.maximum = maximum
        self.exclusiveMinimum = exclusiveMinimum
        self.exclusiveMaximum = exclusiveMaximum
        self.multipleOf = multipleOf


class Integer(NumericElement[int]):
    """JSON Schema ``"integer"`` element.

    Accepts only python `int`.
    """

    @property
    def type_validator(self):
        return InstanceOf(int)


class Number(NumericElement[float]):
    """JSON Schema ``"number"`` element.

    Accepts python ``int`` or ``float``.
    """

    def construct(self, value, _property):  # pylint: disable=no-self-use
        return float(value)

    @property
    def type_validator(self):
        return InstanceOf(float, int)



from typing import Dict

from statham.schema.constants import JSONElement


class StathamError(Exception):
    """Base exception for errors relating to :mod:`statham`."""


class SchemaDefinitionError(StathamError):
    """Raised when invalid schemas are declared in model definitions."""

    @classmethod
    def reserved_attribute(cls, attribute_name: str) -> "SchemaDefinitionError":
        return cls(
            f"May not use reserved attribute `{attribute_name}` as a property "
            "attribute name. Instead use "
            f"`{attribute_name}_ = Property(<element>, "
            f"source='{attribute_name}'`"
        )


class ValidationError(StathamError):
    """Raised when JSON Schema validation fails for input data."""

    @classmethod
    def from_validator(cls, property_, value, message) -> "ValidationError":
        value_string = (
            f"{repr(property_.parent)}.{property_.name} = {repr(value)}`"
            if property_.name != "<unbound>"
            else repr(value)
        )
        return cls(f"Failed validating `{value_string}`. {message}")

    @classmethod
    def combine(
        cls, property_, value, exceptions, message
    ) -> "ValidationError":
        base_message = str(cls.from_validator(property_, value, ""))
        error_breakdown = ", ".join(str(exc) for exc in exceptions)
        error_breakdown = error_breakdown.replace(base_message, "")
        return cls.from_validator(
            property_, value, message + f" Individual errors: {error_breakdown}"
        )

    @classmethod
    def multiple_composition_match(cls, matching_models, data):
        return cls(
            "Matches multiple possible models. Must only match one.\n"
            f"Data: {data}\n"
            f"Models: {matching_models}"
        )


class SchemaParseError(StathamError):
    """Raised when parsing JSON Schema documents to statham models."""

    @classmethod
    def missing_title(
        cls, schema: Dict[str, JSONElement]
    ) -> "SchemaParseError":
        return cls(
            "No title defined in schema. Use "
            "`statham.titles.title_labeller` to pre-process the "
            f"schema: {schema}"
        )

    @classmethod
    def unresolvable_declaration(cls) -> "SchemaParseError":
        return cls(
            "Schema document has an unresolvable declaration tree. This "
            "generally occurs due to cyclical references."
        )

    @classmethod
    def invalid_type(cls, value):
        return cls(f"Got invalid type keyword: {value}.")


# pylint: disable=line-too-long
class FeatureNotImplementedError(SchemaParseError):
    """Raised when parsing valid JSON Schema features currently unsupported by ``statham``."""

    @classmethod
    def unsupported_keywords(cls, keywords) -> "FeatureNotImplementedError":
        return cls(
            f"The following provided keywords are not supported: {keywords}"
        )



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from codecs import open
from os import path

from setuptools import find_packages, setup

import statham


REQUIREMENTS_FILE_PATH = path.join(
    path.abspath(path.dirname(__file__)), "requirements.txt"
)

with open(REQUIREMENTS_FILE_PATH, "r") as f:
    REQUIREMENTS_FILE = [
        line
        for line in f.read().splitlines()
        if not line.startswith("#") and not line.startswith("--")
    ]

with open("README.md", "r") as readme_file:
    LONG_DESCRIPTION = readme_file.read()

setup(
    name="statham-schema",
    version=statham.__version__,
    description=(
        "Tools for generating Python models from JSON Schema documents."
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/jacksmith15/statham-schema",
    author="Jack Smith",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(exclude=["docs*", "tests*"]),
    package_data={"statham": ["py.typed"]},
    install_requires=REQUIREMENTS_FILE,
    dependency_links=[],
    entry_points={"console_scripts": ["statham=statham.__main__:entry_point"]},
)



from typing import Any, Dict, Generic, Optional, TypeVar, TYPE_CHECKING

from statham.schema.constants import NotPassed
from statham.schema.exceptions import SchemaDefinitionError
from statham.schema.helpers import custom_repr_args

if TYPE_CHECKING:
    from statham.schema.elements.base import Element  # pragma: no cover


PropType = TypeVar("PropType")


class _Property(Generic[PropType]):
    """Descriptor for a property on an object.

    Used to bind information about the enclosing object to a target Element,
    e.g. "required". Should be instantiated via `Property` to enable
    type inference on instances.
    """

    required: bool
    parent: Any
    element: "Element"
    source: Optional[str]
    name: Optional[str]

    def __init__(
        self,
        element: "Element[PropType]",
        *,
        required: bool = False,
        source: str = None,
    ):
        self.element = element
        self.required = required
        self.name: Optional[str] = None
        self.source: Optional[str] = source
        self.parent = None

    def __eq__(self, other):
        if not isinstance(other, _Property):
            return False
        return (
            self.element == other.element
            and self.required == other.required
            and self.source == other.source
        )

    def clone(self):
        return _Property(
            self.element, required=self.required, source=self.source
        )

    def evolve(self, name: str) -> "_Property":
        """Generate renamed property object to pass into nested elements."""
        property_: _Property[PropType] = _Property(
            element=self.element, required=self.required, source=self.source
        )
        property_.bind(name=name, parent=self.parent)
        return property_

    def bind(self, name: str = None, parent: "Element" = None) -> None:
        if parent:
            self.parent = parent
        if not name:
            return
        if not self.source:
            self.source = name
        self.name = name

    def __call__(self, value):
        return self.element(value, self)

    def __repr__(self):
        repr_args = custom_repr_args(self)
        if self.source == self.name:
            _ = repr_args.kwargs.pop("source", None)
        return f"{self.__class__.__name__.lstrip('_')}{repr(repr_args)}"

    @property
    def annotation(self):
        if self.required or not isinstance(
            getattr(self.element, "default", NotPassed()), NotPassed
        ):
            return self.element.annotation
        return f"Maybe[{self.element.annotation}]"

    def python(self) -> str:
        prop_def = repr(self)
        return (
            (f"{self.name}: {self.annotation} = {prop_def}")
            if self.name
            else prop_def
        )


# Behaves as a wrapper for the `_Property` class.
# pylint: disable=invalid-name
def Property(element: "Element", *, required: bool = False, source: str = None):
    """Descriptor for adding a property when declaring an object schema model.

    Return value is typed to inform instance-level interface (see type stubs).

    :param element: The JSON Schema Element object accepted by this property.
    :param required: Whether this property is required. If false, then this
        field may be omitted when data is passed to the outer object's
        constructor.
    :param source: The source name of this property. Only necessary if it must
        differ from that of the attribute.
    :return: The property descriptor for this element.

    To hand property name conflicts, use the :paramref:`Property.source`
    option. For example, to express a property called `class`, one could
    do the following:

    .. code:: python

        class MyObject(Object):
            # Property called class
            class_: str = Property(String(), source="class")
    """
    return _Property(element, required=required, source=source)


class _PropertyDict(Dict[str, _Property[Any]]):
    """Container for properties.

    Used internally to bind properties to the enclosing element, and
    attribute name.
    """

    _parent: "Element"

    def __init__(self, *args, **kwargs):
        self._parent = None
        super().__init__(*args, **kwargs)
        bad_values = {
            name: prop
            for name, prop in self.items()  # pylint: disable=no-member
            if not isinstance(prop, _Property)
        }
        if bad_values:
            raise SchemaDefinitionError(f"Got bad property types: {bad_values}")

    def __setitem__(self, key, value):
        if not isinstance(value, _Property):
            raise SchemaDefinitionError(
                f"{key} must be a `Property`, got {value}"
            )
        super().__setitem__(key, value)  # pylint: disable=no-member
        value.bind(name=key, parent=self.parent)

    @property
    def parent(self) -> "Element":
        return self._parent

    @parent.setter
    def parent(self, value: "Element"):
        self._parent = value
        for key, prop in self.items():  # pylint: disable=no-member
            prop.bind(name=key, parent=value)

    @property
    def required(self):
        # pylint: disable=no-member
        return [
            prop.source or name
            for name, prop in self.items()
            if prop.required and isinstance(prop.element.default, NotPassed)
        ]



# False positive. The cycle exists but is avoided by importing last.
# pylint: disable=cyclic-import
from typing import Any, cast, Dict, List, Generic, TypeVar, Union

from statham.schema.constants import NotPassed, Maybe
from statham.schema.exceptions import ValidationError
from statham.schema.helpers import custom_repr
from statham.schema.property import _Property, _PropertyDict
from statham.schema.validation import (
    get_validators,
    InstanceOf,
    NoMatch,
    Validator,
)


T = TypeVar("T")

Numeric = Union[int, float]


# This emulates the options available to a general JSON Schema object.
# pylint: disable=too-many-instance-attributes
class Element(Generic[T]):
    # pylint: disable=line-too-long
    """An un-typed schema element.

    Accepts JSON Schema keywords as arguments.

    :param default:
        The default value this element should return when not provided.
    :param const:
        Restrict this element to a constant value.
    :param enum:
        Restrict this element to an enumeration of provided values.
    :param items:
        The :class:`Element` with which to validate list items. If
        :paramref:`~Element.items` is a list of :class:`Element`, then
        each :class:`Element` validates the corresponding index of the
        value. Subsequent values are validated by
        :paramref:`~Element.additionalItems`.
    :param additionalItems:
        The :class:`Element` with which to validate additional items in a
        list, when :paramref:`items` is a list and shorter than the
        passed list.
    :param minItems:
        The minimum number of items to allow in list values.
    :param maxItems:
        The maximum number of items to allow in list values.
    :param uniqueItems:
        If :const:`True`, validate that list values contain unique
        elements.
    :param contains:
        Validate that list values contain at least one element
        matching this :class:`Element`.
    :param minimum:
        Validate that numeric values are greater than or equal
        to this value.
    :param maximum:
        Validate that numeric values are less than or equal
        to this value.
    :param exclusiveMinimum:
        Validate that numeric values are strictly greater than
        this value.
    :param exclusiveMaximum:
        Validate that numeric values are strictly less than
        this value.
    :param multipleOf:
        Validate that numeric values are are multiple of this
        value.
    :param format:
        Validate that string values conform to the specified
        format. See :func:`~statham.schema.validation.format.format_checker`
        to learn about included formats and how to add your own.
    :param pattern:
        Validate that string values match this value as a regular
        expression.
    :param minLength:
        Validate that string values are at least as long as this
        value.
    :param maxLength:
        Validate that string values are at most as long as this
        value.
    :param required:
        Validate that dictionary values contain each of these
        keys.
    :param properties:
        Validate that dictionary values match these properties.
        See :class:`statham.schema.property.Property` for more information.
    :param patternProperties:
        Validate that keys of dictionaries matching these
        patterns conform to the provided :class:`Element`. Keys may match
        on multiple :paramref:`~Element.patternProperties` and
        :paramref:`~Element.properties` at the same time.
    :param additionalProperties:
        Validate properties not matched by either
        :paramref:`~Element.properties` or
        :paramref:`~Element.patternProperties` against this element. If
        :const:`True`, any value is allowed. If :const:`False`, no
        additional properties are allowed.
    :param minProperties:
        Validate that dictionary values contain at least this
        many members.
    :param maxProperties:
        Validate that dictionary values contain at most this
        many members.
    :param propertyNames:
        Validate that keys of dictionary values match are accepted
        by this :class:`Element`. Any :class:`Element` is allowed here,
        but it is not useful to use any other than
        :class:`statham.schema.elements.String`.
    :param dependencies:
        Define JSON Schema dependencies. There are two types of
        dependencies. If the dict value is a list, then the each property
        name in that list must be present when the property name of the
        key is present. If the value is an :class:`Element`, then it's
        validation applies whenever the property name of the key is present. See
        `Object Dependencies <https://json-schema.org/understanding-json-schema/reference/object.html#dependencies>`_
        for more detail.
    """
    # pylint: enable=line-too-long

    _properties: Maybe[_PropertyDict]

    # This is how many options there are!
    # pylint: disable=too-many-locals
    def __init__(
        self,
        *,
        # Bad name to match JSON Schema keywords.
        # pylint: disable=redefined-builtin
        default: Maybe[Any] = NotPassed(),
        const: Maybe[Any] = NotPassed(),
        enum: Maybe[List[Any]] = NotPassed(),
        items: Maybe[Union["Element", List["Element"]]] = NotPassed(),
        additionalItems: Union["Element", bool] = True,
        minItems: Maybe[int] = NotPassed(),
        maxItems: Maybe[int] = NotPassed(),
        uniqueItems: bool = False,
        contains: Maybe["Element"] = NotPassed(),
        minimum: Maybe[Numeric] = NotPassed(),
        maximum: Maybe[Numeric] = NotPassed(),
        exclusiveMinimum: Maybe[Numeric] = NotPassed(),
        exclusiveMaximum: Maybe[Numeric] = NotPassed(),
        multipleOf: Maybe[Numeric] = NotPassed(),
        format: Maybe[str] = NotPassed(),
        pattern: Maybe[str] = NotPassed(),
        minLength: Maybe[int] = NotPassed(),
        maxLength: Maybe[int] = NotPassed(),
        required: Maybe[List[str]] = NotPassed(),
        properties: Maybe[Dict[str, "_Property"]] = NotPassed(),
        patternProperties: Maybe[Dict[str, "Element"]] = NotPassed(),
        additionalProperties: Union["Element", bool] = True,
        minProperties: Maybe[int] = NotPassed(),
        maxProperties: Maybe[int] = NotPassed(),
        propertyNames: Maybe["Element"] = NotPassed(),
        dependencies: Maybe[
            Dict[str, Union[List[str], "Element"]]
        ] = NotPassed(),
    ):
        self.default = default
        self.const = const
        self.enum = enum
        self.items = items
        self.additionalItems = additionalItems
        self.minItems = minItems
        self.maxItems = maxItems
        self.uniqueItems = uniqueItems
        self.contains = contains
        self.minimum = minimum
        self.maximum = maximum
        self.exclusiveMinimum = exclusiveMinimum
        self.exclusiveMaximum = exclusiveMaximum
        self.multipleOf = multipleOf
        self.format = format
        self.pattern = pattern
        self.minLength = minLength
        self.maxLength = maxLength
        self.required = required
        # https://github.com/python/mypy/issues/3004
        self.properties = properties  # type: ignore
        self.patternProperties = patternProperties
        self.additionalProperties = additionalProperties
        self.minProperties = minProperties
        self.maxProperties = maxProperties
        self.propertyNames = propertyNames
        self.dependencies = dependencies

    @property
    def properties(self) -> Maybe[_PropertyDict]:
        return self._properties

    @properties.setter
    def properties(self, value: Maybe[Dict[str, _Property]]):
        if isinstance(value, NotPassed):
            self._properties = value
            return
        self._properties = _PropertyDict(cast(Dict[str, _Property], value))
        self._properties.parent = self

    def __repr__(self):
        """Dynamically construct the repr to match value instantiation."""
        return custom_repr(self)

    def __eq__(self, other):
        """Check if two elements are equivalent.

        This just checks the type and public attribute values. Useful for
        testing parser logic, and could be used to automatically DRY up
        messy schemas in future.
        """
        if not isinstance(other, self.__class__):
            return False
        pub_vars = lambda x: {
            k: v
            for k, v in vars(x).items()
            if not k.startswith("_") or k == "_properties"
        }
        return pub_vars(self) == pub_vars(other)

    @property
    def annotation(self) -> str:
        # pylint: disable=no-member
        generic = type(self).__orig_bases__[0].__args__[0]  # type: ignore
        if isinstance(generic, TypeVar):  # type: ignore
            return "Any"
        return generic.__name__

    @property
    def type_validator(self) -> Validator:
        return InstanceOf()

    @property
    def validators(self) -> List[Validator]:
        validators: List[Validator] = [self.type_validator] + list(
            get_validators(self)
        )
        return validators

    def construct(self, value, property_):
        if isinstance(value, list):
            return self.__items__(value, property_)
        if isinstance(value, dict):
            return _AnonymousObject(**self.__properties__(value))
        return value

    @property
    def __properties__(self) -> "Properties":
        return Properties(
            self,
            getattr(self, "properties", NotPassed()),
            getattr(self, "patternProperties", NotPassed()),
            getattr(self, "additionalProperties", True),
        )

    @property
    def __items__(self) -> "Items":
        return Items(
            getattr(self, "items", NotPassed()),
            getattr(self, "additionalItems", True),
        )

    def __call__(self, value: Any, property_=None) -> Maybe[T]:
        """Validate and convert input data against this :class:`Element`.

        Dictionary (object) values which pass validation are returned as
        a subclass of :const:`dict` allowing attribute access.

        :param value: The input data.
        :param property_: Optionally specify the outer property scope
            enclosing this :class:`Element`. Used automatically by object
            validation to produce more useful error messages.
        :return: The parsed value.
        """
        property_ = property_ or UNBOUND_PROPERTY

        def create(value):
            for validator in self.validators:
                validator(value, property_)
            return self.construct(value, property_)

        if not isinstance(self.default, NotPassed) and isinstance(
            value, NotPassed
        ):
            try:
                return create(self.default)
            except (TypeError, ValidationError):
                return self.default
        if isinstance(value, NotPassed):
            return value
        return create(value)


UNBOUND_PROPERTY: _Property = _Property(Element(), required=False)
UNBOUND_PROPERTY.bind(parent=Element(), name="<unbound>")


class Nothing(Element):
    """Element which matches nothing. Equivalent to :const:`False`."""

    def __init__(self):
        super().__init__()  # Don't allow args to Nothing.

    def __bool__(self):
        return False

    @property
    def annotation(self) -> str:
        return "None"

    @property
    def validators(self) -> List[Validator]:
        return [NoMatch()]


# Needs to be imported last to prevent cyclic import.
# pylint: disable=wrong-import-position
from statham.schema.elements.properties import Properties
from statham.schema.elements.items import Items


class _AnonymousObject(dict):
    """Anonymous object type.

    Allows dictionaries passed to untyped objects to use attribute access,
    to match `Object` instances.
    """

    def __getattr__(self, key: str) -> Any:
        return self.__getitem__(key)

    def __setattr__(self, key: str, value: Any):
        return self.__setitem__(key, value)



from statham.schema.constants import NotPassed
from statham.schema.elements.base import Element, Nothing


class Items:
    """Interface for retrieving relevant schemas given an array index.

    Used internally by :class:`~statham.schema.elements.Element`.
    """

    def __init__(self, items, additional=True):
        self.items = Element() if isinstance(items, NotPassed) else items
        if isinstance(additional, bool):
            self.additional = {True: Element(), False: Nothing()}[additional]
        else:
            self.additional = additional

    @staticmethod
    def property(property_, index):
        return property_.evolve(name=property_.name or "" + f"[{index}]")

    def __repr__(self):
        items = [repr(self.items)]
        if self.additional == Nothing():
            items.append("additionalItems=False")
        elif self.additional != Element():
            items.append(f"additionalItems={self.additional}")
        return f"{type(self).__name__}({', '.join(items)})"

    def __getitem__(self, index):
        if not isinstance(self.items, list):
            return self.items
        try:
            return self.items[index]
        except IndexError:
            return self.additional

    def __call__(self, value, property_):
        return [
            self[index](sub_value, self.property(property_, index))
            for index, sub_value in enumerate(value)
        ]



"""Iteration tools for Element trees."""
from itertools import chain
from typing import Any, Dict, Iterator, List, Set

from statham.schema.elements import Element
from statham.schema.elements.meta import ObjectMeta
from statham.schema.exceptions import SchemaParseError


# TODO: Generate non-object elements.


def orderer(*elements: Element) -> Iterator[ObjectMeta]:
    """Iterate object classes in declaration order.

    Inspects the full DAG of elements and orders in such a way that
    each object class' dependencies are returned before it is.

    Accepts multiple elements, in case there are multiple start points of
    the DAG.

    Assumes each object class in the tree has a unique name.
    """
    object_classes: List[ObjectMeta] = get_object_classes(*elements)
    object_dependencies: Dict[str, List[str]] = {
        object_class.__name__: [
            dep.__name__
            for dep in get_children(object_class)
            if isinstance(dep, ObjectMeta)
        ]
        for object_class in object_classes
    }

    def has_cycle(name):
        return name in object_dependencies[name]

    def from_name(name: str) -> ObjectMeta:
        """Get the object class from the name."""
        return next(filter(lambda oc: oc.__name__ == name, object_classes))

    def pop_name(name: str) -> ObjectMeta:
        """Remove a object class from each sub-tree, and return it."""
        object_dependencies.update(
            {
                name_: [dep for dep in deps if dep != name]
                for name_, deps in object_dependencies.items()
            }
        )
        del object_dependencies[name]
        return from_name(name)

    cycles: List[str] = sorted(
        name for name in object_dependencies if has_cycle(name)
    )
    if cycles:
        raise SchemaParseError.unresolvable_declaration()

    def _next() -> ObjectMeta:
        next_name = next(
            map(
                lambda _: _[0],
                filter(lambda _: not _[1], object_dependencies.items()),
            )
        )
        return pop_name(next_name)

    try:
        while True:
            yield _next()
    except StopIteration:
        assert not object_dependencies.values()
        return


def get_object_classes(*elements: Element) -> List[ObjectMeta]:
    return [
        element
        for element in list(elements)
        + [child for element in elements for child in get_children(element)]
        if isinstance(element, ObjectMeta)
    ]


def get_children(element: Any, seen: Set[int] = None) -> Iterator[Element]:
    """Iterate all child elements of an element."""
    seen = seen or set()
    if id(element) in seen:
        yield element
        return
    seen.add(id(element))
    paths = [  # Possible paths to immediate children.
        "items",
        "additionalItems",
        "contains",
        "properties.*.element",
        "additionalProperties",
        "patternProperties.*",
        "propertyNames",
        "dependencies.*",
        "elements",
        "element",
    ]
    children = [
        child
        for path in paths
        for child in _get_path(element, path)
        if isinstance(child, Element)
    ]
    for child in children:
        yield child
        yield from get_children(child, seen)


def _get_path(element, path):
    """Get items matching a dot-separated path.

    Assumes dictionary lookup, but falls back to attribute access. If an
    item is a list, chain the remaining results for each item. If a path
    segment is * then get for all values (dictionary only).

    If there is no match on a path, then return nothing at all.
    """
    first, *rest = path.split(".")
    if first == "*":
        try:
            next_item = list(element.values())
        except AttributeError:
            return []
    else:
        try:
            next_item = element[first]
        except (KeyError, TypeError):
            try:
                next_item = getattr(element, first)
            except AttributeError:
                return []
    if isinstance(next_item, list):
        if not rest:
            return next_item
        return list(
            chain.from_iterable(
                _get_path(next_subitem, ".".join(rest))
                for next_subitem in next_item
            )
        )
    if not rest:
        return [next_item]
    return _get_path(next_item, ".".join(rest))
